from typing import Dict

from officialeye.config import Config
from officialeye.error.errors.template import ErrTemplateInvalidInterpretation


class InterpretationMethodConfig(Config):

    def __init__(self, config_dict: Dict[str, any], interpretation_method: str, /):

        super().__init__(config_dict)

        self._interpretation_method = interpretation_method

    def _get_invalid_key_error(self, key: str, /):

        return ErrTemplateInvalidInterpretation(
            f"while reading configuration of the '{self._interpretation_method}' interpretation method.",
            f"Could not find a value for key '{key}'."
        )



from queue import Queue
from threading import Thread
from typing import List, Union, Tuple

# noinspection PyPackageRequirements
import cv2

from officialeye.context.singleton import oe_context
from officialeye.error import OEError
from officialeye.error.errors.supervision import ErrSupervisionCorrespondenceNotFound
from officialeye.error.errors.template import ErrTemplateInvalidConcurrencyConfig
from officialeye.error.printing import oe_error_print_debug
from officialeye.supervision.result import SupervisionResult
from officialeye.template.template import Template
from officialeye.util.logger import oe_debug


class AnalysisWorker(Thread):

    def __init__(self, worker_id: int, queue: Queue, target: cv2.Mat, /):
        Thread.__init__(self)
        self.worker_id = worker_id
        self.queue = queue
        self._target = target
        self._results: List[Tuple[Union[SupervisionResult, None], Union[OEError, None]]] = []

    def run(self):

        while True:
            template = self.queue.get()

            try:
                _current_result = template.analyze(self._target)
                self._results.append((_current_result, None))
            except OEError as err:
                self._results.append((None, err))
            finally:
                self.queue.task_done()

    def get_successful_results(self):
        for result, error in self._results:
            if result is None or error is not None:
                continue
            yield result

    def get_errors(self):
        for _, error in self._results:
            if error is not None:
                yield error




def do_analyze(target: cv2.Mat, templates: List[Template], /, *, num_workers: int):

    if len(templates) == 0:
        # the program should be a noop if there are no templates provided
        return

    assert num_workers is not None

    if num_workers < 1:
        raise ErrTemplateInvalidConcurrencyConfig(
            "while setting up workers for analyzing the target image.",
            f"The provided number of workers ({num_workers}) cannot be less than one."
        )

    if num_workers > 0xff:
        raise ErrTemplateInvalidConcurrencyConfig(
            "while setting up workers for analyzing the target image.",
            f"The provided number of workers ({num_workers}) is too high."
        )

    queue = Queue(maxsize=len(templates))

    workers = [AnalysisWorker(worker_id, queue, target) for worker_id in range(num_workers)]

    for worker in workers:
        worker.daemon = True
        worker.start()

    for template in templates:
        queue.put(template)

    queue.join()

    best_result = None
    best_result_score = -1.0

    # a list containing regular errors that occurred in workers
    regular_errors = []

    for worker in workers:
        for result in worker.get_successful_results():
            assert result is not None

            result_score = result.get_score()
            if result_score > best_result_score:
                best_result_score = result_score
                best_result = result

        for error in worker.get_errors():
            assert error is not None

            # we ignore regular errors here, because they may well be simply caused by trying to match
            # a given document against an invalid template, which is normal behavior
            if not error.is_regular:
                raise error
            else:
                oe_debug(f"Worker {worker.worker_id} returned the following non-regular error {error.code_text}:")
                oe_error_print_debug(error)
                regular_errors.append(error)

    # note: best_result may be None here

    if best_result is None:

        error = ErrSupervisionCorrespondenceNotFound(
            "while running supervisor",
            "could not establish correspondence of the image with any of the templates provided"
        )

        for worker_error in regular_errors:
            error.add_cause(worker_error)

        raise error

    oe_context().io_driver.output_supervision_result(target, best_result)



# noinspection PyPackageRequirements
import cv2

from officialeye.context.singleton import oe_context
from officialeye.error import OEError
from officialeye.error.printing import oe_error_print_error
from officialeye.io.driver import IODriver
from officialeye.supervision.result import SupervisionResult
from officialeye.template.template import Template
from officialeye.util.cli_utils import export_and_show_image, export_image


class TestIODriver(IODriver):

    DRIVER_ID = "test"

    def __init__(self):
        super().__init__()

        self.visualize_features: bool = True

    def output_supervision_result(self, target: cv2.Mat, result: SupervisionResult, /):

        assert result is not None

        template = oe_context().get_template(result.template_id)

        application_image = template.load_source_image()

        # extract the features from the target image
        for feature in template.features():
            feature_img = result.get_feature_warped_region(target, feature)
            feature_img_mutated = feature.apply_mutators_to_image(feature_img)
            feature.insert_into_image(application_image, feature_img_mutated)

        if self.visualize_features:
            # visualize features on the image
            for feature in template.features():
                application_image = feature.visualize(application_image)

        if oe_context().quiet_mode:
            export_image(application_image, file_name="supervision_result.png")
        else:
            export_and_show_image(application_image, file_name="supervision_result.png")

    def output_show_result(self, template: Template, img: cv2.Mat, /):
        if oe_context().quiet_mode:
            export_image(img, file_name=f"{template.template_id}.png")
        else:
            export_and_show_image(img, file_name=f"{template.template_id}.png")

    def output_error(self, error: OEError, /):
        oe_error_print_error(error)



import strictyaml as yml

from officialeye.error.errors.template import ErrTemplateInvalidSyntax
from officialeye.template.parser.schema import generate_template_schema
from officialeye.template.template import Template
from officialeye.util.logger import oe_error, oe_info

_oe_template_schema = generate_template_schema()


def _print_error_message(err: yml.StrictYAMLError, template_path: str):

    oe_error("Error ", bold=True, nl=False)

    if err.context is not None:
        oe_error(err.context, prefix=False)
    else:
        oe_error("while parsing", prefix=False)

    if err.context_mark is not None and (
            err.problem is None
            or err.problem_mark is None
            or err.context_mark.name != err.problem_mark.name
            or err.context_mark.line != err.problem_mark.line
            or err.context_mark.column != err.problem_mark.column
    ):
        oe_error(str(err.context_mark).replace("<unicode string>", template_path))

    if err.problem is not None:
        oe_error("Problem", bold=True, nl=False)
        oe_error(f": {err.problem}", prefix=False)

    if err.problem_mark is not None:
        oe_error(str(err.problem_mark).replace("<unicode string>", template_path))


def load_template(path: str) -> Template:
    """
    Loads a template from a file located at the specified path.

    Arguments:
        path: The path to the YAML template configuration file.

    Returns:
        Loaded template.

    Raises:
        OEError: In case there has been an error validating the correctness of the template.
    """

    global _oe_template_schema

    with open(path, "r") as fh:
        raw_data = fh.read()

    try:
        yaml_document = yml.load(raw_data, schema=_oe_template_schema)
    except yml.StrictYAMLError as err:
        _print_error_message(err, path)
        exit(4)
    except yml.YAMLError:
        raise ErrTemplateInvalidSyntax(
            f"while loading template configuration file at '{path}'.",
            f"General parsing error. Check the syntax and the encoding of the file."
        )

    data = yaml_document.data

    template = Template(data, path)

    oe_info(f"Loaded template: ", nl=False)
    oe_info(str(template), prefix=False, bold=True)

    return template



from typing import List

# noinspection PyPackageRequirements
import cv2

from officialeye.util.cli_utils import export_image


class DebugImage:

    def __init__(self, img: cv2.Mat, name: str, /):
        self.img = img
        self.name = name

    def get_export_filename(self, image_id: int) -> str:

        if self.name == "":
            return "%03d.png" % image_id

        return "%03d_%s.png" % (image_id, self.name)


class DebugContainer:

    def __init__(self):
        self._images: List[DebugImage] = []

    def add_image(self, img: cv2.Mat, /, *, name: str = ""):
        self._images.append(DebugImage(img, name))

    def export(self):

        for image_id, image in enumerate(self._images):
            export_filename = image.get_export_filename(image_id)
            export_image(image.img, file_name=export_filename)



import click
# noinspection PyPackageRequirements
import cv2

from officialeye.context.singleton import oe_context
from officialeye.util.logger import oe_info


def export_image(img: cv2.Mat, /, *, file_name: str = "") -> str:
    export_file_path = oe_context().allocate_file_for_export(file_name=file_name)
    cv2.imwrite(export_file_path, img)
    oe_info(f"Exported '{export_file_path}'.")
    return export_file_path


def export_and_show_image(img: cv2.Mat, /, *, file_name: str = ""):
    path = export_image(img, file_name=file_name)
    click.launch(path, locate=False)
    click.pause()



from typing import List

import click
# noinspection PyPackageRequirements
import cv2

from officialeye.context.singleton import oe_context
from officialeye.error import OEError
from officialeye.io.drivers.run import RunIODriver
from officialeye.io.drivers.test import TestIODriver
from officialeye.meta import OFFICIALEYE_GITHUB, OFFICIALEYE_VERSION, print_logo
from officialeye.template.analyze import do_analyze
from officialeye.template.create import create_example_template_config_file
from officialeye.template.parser.loader import load_template
from officialeye.util.logger import oe_info, oe_warn


@click.group()
@click.option("-d", "--debug", is_flag=True, show_default=True, default=False, help="Enable debug mode.")
@click.option("--dedir", type=click.Path(exists=True, file_okay=True, readable=True), help="Specify debug export directory.")
@click.option("--edir", type=click.Path(exists=True, file_okay=True, readable=True), help="Specify export directory.")
@click.option("-q", "--quiet", is_flag=True, show_default=True, default=False, help="Disable standard output messages.")
@click.option("-v", "--verbose", is_flag=True, show_default=True, default=False, help="Enable verbose logging.")
@click.option("-dl", "--disable-logo", is_flag=True, show_default=True, default=False, help="Disable the officialeye logo.")
def cli(debug: bool, dedir: str, edir: str, quiet: bool, verbose: bool, disable_logo: bool):

    oe_context().debug_mode = debug
    oe_context().quiet_mode = quiet
    oe_context().verbose_mode = verbose
    oe_context().disable_logo = disable_logo

    oe_context().io_driver = TestIODriver()

    if not quiet and not disable_logo:
        print_logo()

    if dedir is not None:
        oe_context().debug_export_directory = dedir

    if edir is not None:
        oe_context().export_directory = edir

    if oe_context().debug_mode:
        oe_warn("Debug mode enabled. Disable for production use to improve performance.")


# noinspection PyShadowingBuiltins
@click.command()
@click.argument("template_path", type=click.Path(exists=False, file_okay=True, readable=True, writable=True))
@click.argument("template_image", type=click.Path(exists=True, file_okay=True, readable=True, writable=False))
@click.option("--id", type=str, show_default=False, default="example", help="Specify the template identifier.")
@click.option("--name", type=str, show_default=False, default="Example", help="Specify the template name.")
@click.option("--force", is_flag=True, show_default=True, default=False, help="Create missing directories and overwrite file.")
def create(template_path: str, template_image: str, id: str, name: str, force: bool):
    """Creates a new template configuration file at the specified path."""

    try:
        create_example_template_config_file(template_path, template_image, id, name, force)
    except OEError as err:
        oe_context().io_driver.output_error(err)
    finally:
        oe_context().dispose()


@click.command()
@click.argument("template_path", type=click.Path(exists=True, file_okay=True, readable=True))
@click.option("--hide-features", is_flag=True, show_default=False, default=False, help="Do not visualize the locations of features.")
@click.option("--hide-keypoints", is_flag=True, show_default=False, default=False, help="Do not visualize the locations of keypoints.")
def show(template_path: str, hide_features: bool, hide_keypoints: bool):
    """Exports template as an image with features visualized."""
    try:
        template = load_template(template_path)
        img = template.show(hide_features=hide_features, hide_keypoints=hide_keypoints)
        oe_context().io_driver.output_show_result(template, img)
    except OEError as err:
        oe_context().io_driver.output_error(err)
    finally:
        oe_context().dispose()


@click.command()
@click.argument("target_path", type=click.Path(exists=True, file_okay=True, readable=True))
@click.argument("template_paths", type=click.Path(exists=True, file_okay=True, readable=True), nargs=-1)
@click.option("--workers", type=int, default=4, show_default=True)
@click.option("--show-features", is_flag=True, show_default=False, default=False, help="Visualize the locations of features.")
def test(target_path: str, template_paths: List[str], workers: int, show_features: bool):
    """Visualizes the analysis of an image using one or more templates."""

    assert isinstance(oe_context().io_driver, TestIODriver)
    oe_context().io_driver.visualize_features = show_features

    # load target image
    target = cv2.imread(target_path, cv2.IMREAD_COLOR)

    try:
        templates = [load_template(template_path) for template_path in template_paths]
        do_analyze(target, templates, num_workers=workers)
    except OEError as err:
        oe_context().io_driver.output_error(err)
    finally:
        oe_context().dispose()


@click.command()
@click.argument("target_path", type=click.Path(exists=True, file_okay=True, readable=True))
@click.argument("template_paths", type=click.Path(exists=True, file_okay=True, readable=True), nargs=-1)
@click.option("--workers", type=int, default=4, show_default=True)
def run(target_path: str, template_paths: List[str], workers: int):
    """Applies one or more templates to an image."""

    # load target image
    target = cv2.imread(target_path, cv2.IMREAD_COLOR)

    # overwrite the IO driver
    oe_context().io_driver = RunIODriver()

    try:
        templates = [load_template(template_path) for template_path in template_paths]
        do_analyze(target, templates, num_workers=workers)
    except OEError as err:
        oe_context().io_driver.output_error(err)
    finally:
        oe_context().dispose()


@click.command()
def homepage():
    """Go to the officialeye's official GitHub homepage."""
    oe_info(f"Opening {OFFICIALEYE_GITHUB}")
    click.launch(OFFICIALEYE_GITHUB)
    oe_context().dispose()


@click.command()
def version():
    """Print the version of OfficialEye."""
    oe_info(f"Version: v{OFFICIALEYE_VERSION}")
    oe_context().dispose()


cli.add_command(create)
cli.add_command(show)
cli.add_command(test)
cli.add_command(run)
cli.add_command(homepage)
cli.add_command(version)

if __name__ == "__main__":
    cli()



from typing import Dict

# noinspection PyPackageRequirements
import cv2
from pytesseract import pytesseract

from officialeye.interpretation import InterpretationMethod


class TesseractInterpretationMethod(InterpretationMethod):

    METHOD_ID = "ocr_tesseract"

    def __init__(self, config_dict: Dict[str, any]):
        super().__init__(TesseractInterpretationMethod.METHOD_ID, config_dict)

        self._tesseract_lang = self.get_config().get("lang", default="eng")
        self._tesseract_config = self.get_config().get("config", default="--dpi 10000 --oem 3 --psm 6")

    def interpret(self, feature_img: cv2.Mat, feature_id: str, /) -> any:
        return pytesseract.image_to_string(feature_img, lang=self._tesseract_lang, config=self._tesseract_config).strip()



# needed to not break type annotations if we are not in type checking mode
from __future__ import annotations

import os
import shutil
import tempfile
from typing import List, Dict
# needed to avoid template.py - singleton.py - context.py - template.py circular import
from typing import TYPE_CHECKING

import click
# noinspection PyPackageRequirements
import z3

from officialeye.error.errors.template import ErrTemplateIdNotUnique
from officialeye.meta import OFFICIALEYE_NAME

if TYPE_CHECKING:
    from officialeye.template.template import Template


class Context:

    def __init__(self):
        self._export_counter = 1
        self._not_deleted_temporary_files: List[str] = []
        self.debug_mode = False
        self.quiet_mode = False
        self.verbose_mode = False
        self.disable_logo = False
        self.debug_export_directory = None
        self.export_directory = None
        self.io_driver = None
        # keys: template ids
        # values: template
        self._loaded_templates: Dict[str, Template] = {}

        z3.set_param("parallel.enable", True)

    def on_template_loaded(self, template: Template, /):

        if template.template_id in self._loaded_templates:
            raise ErrTemplateIdNotUnique(
                f"while loading template '{template.template_id}'",
                "A template with the same id has already been loaded."
            )

        self._loaded_templates[template.template_id] = template

    def get_template(self, template_id: str, /) -> Template:
        assert template_id in self._loaded_templates, "Unknown template id"
        return self._loaded_templates[template_id]

    def get_debug_export_directory(self) -> str:
        if self.debug_export_directory is not None:
            return self.debug_export_directory

        debug_directory = os.path.join(click.get_app_dir(OFFICIALEYE_NAME), "debug")

        if os.path.exists(debug_directory):
            shutil.rmtree(debug_directory)

        os.makedirs(debug_directory, exist_ok=True)

        # cache the debug export directory
        self.debug_export_directory = debug_directory

        return self.debug_export_directory

    def _allocate_file_name(self) -> str:
        file_name = "%03d.png" % self._export_counter
        self._export_counter += 1
        return file_name

    def _allocate_file_for_debug_export(self, file_name: str = "") -> str:
        assert self.debug_mode, "Tried to export debug file when debug mode is off"

        if file_name == "":
            file_name = self._allocate_file_name()

        return os.path.join(self.get_debug_export_directory(), file_name)

    def allocate_file_for_export(self, /, *, file_name: str = "") -> str:

        if self.debug_mode:
            return self._allocate_file_for_debug_export(file_name=file_name)

        if self.export_directory is None:
            with tempfile.NamedTemporaryFile(prefix="officialeye_", suffix=".png", delete=False) as fp:
                fp.close()
            self._not_deleted_temporary_files.append(fp.name)
            return fp.name

        if file_name == "":
            file_name = self._allocate_file_name()

        return os.path.join(self.export_directory, file_name)

    def _cleanup_temporary_files(self):
        while len(self._not_deleted_temporary_files) > 0:
            temp_file = self._not_deleted_temporary_files.pop()
            if os.path.exists(temp_file):
                os.unlink(temp_file)

    def dispose(self):
        self._cleanup_temporary_files()



import click

from officialeye.context.singleton import oe_context


def oe_info(msg, *args, prefix: bool = True, **kwargs):
    if oe_context().quiet_mode:
        return
    if prefix:
        click.secho("INFO", bold=True, bg="blue", nl=False)
        click.echo("  ", nl=False)
    click.secho(msg, *args, **kwargs)


def oe_warn(msg, *args, prefix: bool = True, **kwargs):
    if oe_context().quiet_mode:
        return
    if prefix:
        click.secho("WARN", bold=True, bg="yellow", nl=False)
        click.echo("  ", nl=False)
    click.secho(msg, *args, **kwargs)


def oe_debug(msg, *args, prefix: bool = True, **kwargs):
    if not oe_context().debug_mode:
        return
    if oe_context().quiet_mode:
        return
    if prefix:
        click.secho("DEBUG", bold=True, bg="yellow", nl=False)
        click.echo(" ", nl=False)
    click.secho(msg, *args, **kwargs)


def oe_debug_verbose(msg, *args, prefix: bool = True, **kwargs):
    if not oe_context().debug_mode:
        return
    if oe_context().quiet_mode:
        return
    if not oe_context().verbose_mode:
        return
    if prefix:
        click.secho("DEBUG", bold=True, bg="yellow", nl=False)
        click.echo(" ", nl=False)
    click.secho(msg, *args, **kwargs)


def oe_error(msg, *args, prefix: bool = True, **kwargs):
    if oe_context().quiet_mode:
        return
    for line in msg.splitlines():
        if prefix:
            click.secho("ERROR", bold=True, bg="red", nl=False, err=True)
            click.echo(" ", nl=False, err=True)
        click.secho(line, *args, **kwargs, err=True)




from typing import Dict

from officialeye.error.errors.template import ErrTemplateInvalidInterpretation
from officialeye.interpretation import InterpretationMethod
from officialeye.interpretation.interpretations.tesseract import TesseractInterpretationMethod


def load_interpretation_method(method_id: str, config_dict: Dict[str, any], /) -> InterpretationMethod:

    if method_id == TesseractInterpretationMethod.METHOD_ID:
        return TesseractInterpretationMethod(config_dict)

    raise ErrTemplateInvalidInterpretation(
        f"while loading interpretation method '{method_id}'.",
        "Unknown interpretation method id."
    )



import cv2


def remove_noise(img):
    # blur
    blur = cv2.GaussianBlur(img, (0, 0), sigmaX=33, sigmaY=33)

    # divide
    divide = cv2.divide(img, blur, scale=255)

    # otsu threshold
    thresh = cv2.threshold(divide, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]

    # apply morphology
    kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (3, 3))
    morph = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel)

    return morph



from officialeye.error import OEError
from officialeye.util.logger import oe_error, oe_debug


def oe_error_print_error(error: OEError):
    oe_error(f"Error {error.code} in module {error.module}: {error.code_text}")
    oe_error(f"Error occurred {error.while_text}")
    oe_error(f"Problem: {error.problem_text}")


def oe_error_print_debug(error: OEError):
    oe_debug(f"Error {error.code} in module {error.module}: {error.code_text}")
    oe_debug(f"Error occurred {error.while_text}")
    oe_debug(f"Problem: {error.problem_text}")



import strictyaml as yml

from officialeye.diffobject.difference_expansion import DiffObjectExpansion
from officialeye.diffobject.difference_modes import DIFF_MODE_ADD, DIFF_MODE_OVERRIDE
from officialeye.diffobject.specification import DiffObjectSpecification
from officialeye.diffobject.specification_entries.integer import IntegerSpecificationEntry
from officialeye.diffobject.specification_entries.list import ListSpecificationEntry
from officialeye.diffobject.specification_entries.string import StringSpecificationEntry


def test_01():
    spec = DiffObjectSpecification({
        "one": StringSpecificationEntry(yml.Any()),
        "two": StringSpecificationEntry(yml.Any()),
        "three": StringSpecificationEntry(yml.Any()),
    })

    expansion = DiffObjectExpansion(spec)

    expansion.add({
        "one": "1"
    })

    assert expansion.get_current_partial_object() == {
        "one": "1"
    }

    expansion.add({
        "two": "2"
    })

    assert expansion.get_current_partial_object() == {
        "one": "1",
        "two": "2"
    }

    expansion.add({
        "three": "3"
    })

    assert expansion.get_full_object() == {
        "one": "1",
        "two": "2",
        "three": "3"
    }


def test_02():
    spec = DiffObjectSpecification({
        "one": StringSpecificationEntry(yml.Any()),
        "two": StringSpecificationEntry(yml.Any()),
        "three": StringSpecificationEntry(yml.Any()),
    })

    expansion = DiffObjectExpansion(spec)

    expansion.add({
        "one": "1"
    })

    expansion.add({
        "two": "2"
    })

    assert expansion.get_current_partial_object() == {
        "one": "1",
        "two": "2"
    }

    expansion.add({
        "three": "3"
    })

    expansion.add({
        "two": "lol",
        "$two": DIFF_MODE_ADD,
        "three": "4",
        "$three": DIFF_MODE_OVERRIDE
    })

    assert expansion.get_full_object() == {
        "one": "1",
        "two": "2lol",
        "three": "4"
    }


def test_03():
    spec = DiffObjectSpecification({
        "one": StringSpecificationEntry(yml.Any()),
        "two": {
            "a": ListSpecificationEntry(yml.Any()),
            "b": IntegerSpecificationEntry(yml.Int())
        }
    })

    expansion = DiffObjectExpansion(spec)

    expansion.add({
        "one": "1",
        "two": {
            "a": [1, 2, 3, 4]
        }
    })

    assert expansion.get_current_partial_object() == {
        "one": "1",
        "two": {
            "a": [1, 2, 3, 4]
        }
    }

    expansion.add({
        "two": {
            "a": [5, 6],
            "$a": DIFF_MODE_ADD,
            "b": 0xff
        }
    })

    assert expansion.get_full_object() == {
        "one": "1",
        "two": {
            "a": [1, 2, 3, 4, 5, 6],
            "b": 0xff
        }
    }



import abc
from abc import ABC

# noinspection PyPackageRequirements
import cv2

from officialeye.error import OEError
from officialeye.supervision.result import SupervisionResult
from officialeye.template.template import Template


class IODriver(ABC):

    def __init__(self):
        pass

    @abc.abstractmethod
    def output_supervision_result(self, target: cv2.Mat, result: SupervisionResult, /):
        raise NotImplementedError()

    @abc.abstractmethod
    def output_show_result(self, template: Template, img: cv2.Mat, /):
        raise NotImplementedError()

    @abc.abstractmethod
    def output_error(self, error: OEError, /):
        raise NotImplementedError()



import abc
from abc import ABC

# noinspection PyPackageRequirements
import cv2

from officialeye.context.singleton import oe_context
from officialeye.debug.debuggable import Debuggable
from officialeye.matching.matcher_config import KeypointMatcherConfig
from officialeye.matching.result import KeypointMatchingResult
from officialeye.util.logger import oe_warn


class KeypointMatcher(ABC, Debuggable):

    def __init__(self, engine_id: str, template_id: str, img: cv2.Mat, /):
        super().__init__()

        self.__engine_id = engine_id
        self.template_id = template_id

        # retreive configurations for all keypoint matching engines
        matching_config = self.get_template().get_matching_config()

        assert isinstance(matching_config, dict)

        # get the configuration for the particular engine of interest
        if self.__engine_id in matching_config:
            config_dict = matching_config[self.__engine_id]
        else:
            oe_warn(f"Could not find any configuration entries for the '{self.__engine_id}' matching engine that is being used.")
            config_dict = {}

        self._config = KeypointMatcherConfig(config_dict, engine_id)

        self._img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    @abc.abstractmethod
    def match_keypoint(self, pattern: cv2.Mat, keypoint_id: str, /):
        raise NotImplementedError()

    @abc.abstractmethod
    def match_finish(self) -> KeypointMatchingResult:
        raise NotImplementedError()

    def get_template(self):
        return oe_context().get_template(self.template_id)

    def get_config(self) -> KeypointMatcherConfig:
        return self._config



from officialeye.context.singleton import oe_context
from officialeye.debug.container import DebugContainer


class Debuggable:

    def __init__(self):
        if oe_context().debug_mode:
            self._debug = DebugContainer()
        else:
            self._debug = None

    def in_debug_mode(self, /) -> bool:
        return self._debug is not None

    def debug(self) -> DebugContainer:
        return self._debug



from typing import Dict

from officialeye.error.errors.template import ErrTemplateInvalidMutator
from officialeye.mutator.mutator import Mutator
from officialeye.mutator.mutators.clahe import CLAHEMutator
from officialeye.mutator.mutators.grayscale import GrayscaleMutator
from officialeye.mutator.mutators.non_local_means_denoising import NonLocalMeansDenoisingMutator


def load_mutator(mutator_id: str, config: Dict[str, any], /) -> Mutator:

    if mutator_id == GrayscaleMutator.MUTATOR_ID:
        return GrayscaleMutator(config)

    if mutator_id == NonLocalMeansDenoisingMutator.MUTATOR_ID:
        return NonLocalMeansDenoisingMutator(config)

    if mutator_id == CLAHEMutator.MUTATOR_ID:
        return CLAHEMutator(config)

    raise ErrTemplateInvalidMutator(
        f"while loading mutator '{mutator_id}'.",
        "Unknown mutator id."
    )


def load_mutator_from_dict(mutator_dict: Dict[str, any], /) -> Mutator:

    assert "id" in mutator_dict

    mutator_id = mutator_dict["id"]

    if "config" in mutator_dict:
        mutator_config = mutator_dict["config"]
    else:
        mutator_config = {}

    return load_mutator(mutator_id, mutator_config)
